######################################################################
#
# $Id: FdaRoutines.pm,v 1.8 2008/05/19 14:45:43 mavrik Exp $
#
######################################################################
#
# Copyright 2007-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: File, directory, and attribute routines
#
######################################################################

package WebJob::FdaRoutines;

require Exporter;

use 5.008;
use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
use WebJob::ValidationRoutines;

@EXPORT = qw(FdaCreateDirectory FdaCreateFile FdaGroupToGid FdaOwnerToUid FdaPermissionsToMode);

@EXPORT_OK = ();
@ISA = qw(Exporter);
$VERSION = do { my @r = (q$Revision: 1.8 $ =~ /(\d+)/g); sprintf("%d."."%03d" x $#r, @r); };

######################################################################
#
# FdaGroupToGid
#
######################################################################

sub FdaGroupToGid
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' => ['Group'],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # If the group is already a number, return it. Otherwise, attempt
  # to resolve the group name.
  #
  ####################################################################

  my $sGid = ($$phPArgs{'Group'} =~ /^\d+$/) ? $$phPArgs{'Group'} : getgrnam($$phPArgs{'Group'});
  if (!defined($sGid))
  {
    $$phPArgs{'Error'} = "Unable to resolve GID for $$phPArgs{'Group'}.";
    return undef;
  }

  return $sGid;
}


######################################################################
#
# FdaCreateDirectory
#
######################################################################

sub FdaCreateDirectory
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
#     'BeQuiet', # Optional
      'Directory',
      'Gid',
      'Mode',
#     'Prefix',  # Optional
      'Uid',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Create directory tree -- one element at a time.
  #
  ####################################################################

  my $sBeQuiet = $$phPArgs{'BeQuiet'} || 0;

  my $sDirectory = $$phPArgs{'Directory'};
  $sDirectory =~ s/^[\/\\]+//; # Remove leading slashes.
  $sDirectory =~ s/^([\/\\])+/$1/g; # Squash inner slashes.
  $sDirectory =~ s/[\/\\]+$//; # Remove trailing slashes.

  my @aPathElements = split(/[\/\\]/, $sDirectory);

  my $sPath = $$phPArgs{'Prefix'} || undef;

  foreach my $sPathElement (@aPathElements)
  {
    if (!defined($sPath))
    {
      $sPath = ($sPathElement =~ /^[A-Za-z]:$/) ? "$sPathElement" : "/$sPathElement";
    }
    else
    {
      $sPath .= "/$sPathElement";
    }
    if (!-d $sPath)
    {
      print "Creating $sPath\n" unless ($$phPArgs{'BeQuiet'});
      if (!mkdir($sPath, $$phPArgs{'Mode'}))
      {
        $$phPArgs{'Error'} = "Directory ($sPath) could not be created ($!).";
        return undef;
      }
    }
    else
    {
      my $sTargetPath;
      if (!defined($$phPArgs{'Prefix'}))
      {
        $sTargetPath = ($sDirectory =~ /^[A-Za-z]:/) ? "$sDirectory" : "/$sDirectory";
      }
      else
      {
        $sTargetPath = $$phPArgs{'Prefix'} . "/$sDirectory";
      }
      next unless ($sPath eq $sTargetPath); # Only update ownership/permissions when the current path equals the target path.
      print "Updating $sPath (already exists)\n" unless ($$phPArgs{'BeQuiet'});
    }
    if (!chmod($$phPArgs{'Mode'}, $sPath))
    {
      $$phPArgs{'Error'} = "Unable to set permissions (" . sprintf("%04o", $$phPArgs{'Mode'}) . ") for $sPath ($!).";
      return undef;
    }
    if (!chown($$phPArgs{'Uid'}, $$phPArgs{'Gid'}, $sPath))
    {
      $$phPArgs{'Error'} = "Unable to set owner/group ($$phPArgs{'Uid'}/$$phPArgs{'Gid'}) for $sPath ($!).";
      return undef;
    }
  }

  1;
}


######################################################################
#
# FdaCreateFile
#
######################################################################

sub FdaCreateFile
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
#     'BeQuiet', # Optional
      'File',
      'Gid',
      'Mode',
#     'Prefix',  # Optional
      'Uid',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Create file.
  #
  ####################################################################

  my $sBeQuiet = $$phPArgs{'BeQuiet'} || 0;

  my $sFile = ($$phPArgs{'Prefix'}) ? $$phPArgs{'Prefix'} . "/" . $$phPArgs{'File'} : $$phPArgs{'File'};

  if (!-f $sFile)
  {
    print "Creating $sFile\n" unless ($$phPArgs{'BeQuiet'});
    if (!open(FDAFH, "> $sFile"))
    {
      $$phPArgs{'Error'} = "File ($sFile) could not be created ($!).";
      return undef;
    }
    close(FDAFH);
  }
  else
  {
    print "Updating $sFile (already exists)\n" unless ($$phPArgs{'BeQuiet'});
  }
  if (!chmod($$phPArgs{'Mode'}, $sFile))
  {
    $$phPArgs{'Error'} = "Unable to set permissions (" . sprintf("%04o", $$phPArgs{'Mode'}) . ") for $sFile ($!).";
    return undef;
  }
  if (!chown($$phPArgs{'Uid'}, $$phPArgs{'Gid'}, $sFile))
  {
    $$phPArgs{'Error'} = "Unable to set owner/group ($$phPArgs{'Uid'}/$$phPArgs{'Gid'}) for $sFile ($!).";
    return undef;
  }

  1;
}


######################################################################
#
# FdaOwnerToUid
#
######################################################################

sub FdaOwnerToUid
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' => ['Owner'],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # If the owner is already a number, return it. Otherwise, attempt
  # to resolve the owner name.
  #
  ####################################################################

  my $sUid = ($$phPArgs{'Owner'} =~ /^\d+$/) ? $$phPArgs{'Owner'} : getpwnam($$phPArgs{'Owner'});
  if (!defined($sUid))
  {
    $$phPArgs{'Error'} = "Unable to resolve UID for $$phPArgs{'Owner'}.";
    return undef;
  }

  return $sUid;
}


######################################################################
#
# FdaPermissionsToMode
#
######################################################################

sub FdaPermissionsToMode
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' => ['Permissions'],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Convert the permissions to octal, and apply the umask if one was
  # specified.
  #
  ####################################################################

  my $sMode = (defined($$phPArgs{'Umask'})) ? oct($$phPArgs{'Permissions'}) & ~($$phPArgs{'Umask'}) : oct($$phPArgs{'Permissions'});

  return $sMode;
}

1;

__END__

=pod

=head1 NAME

WebJob::FdaRoutines - File, directory, and attribute routines

=head1 SYNOPSIS

    use WebJob::FdaRoutines;

=head1 DESCRIPTION

This module is a collection of file, directory, and attribute routines
designed to support various WebJob server-side utilities. As such,
minimal effort was put into supporting this code for general
consumption. In other words, use at your own risk and don't expect the
interface to remain the same or backwards compatible from release to
release. This module does not provide an OO interface, nor will it do
so anytime soon.

=head1 AUTHOR

Klayton Monroe

=head1 LICENSE

All documentation and code are distributed under same terms and
conditions as B<WebJob>.

=cut
