######################################################################
#
# $Id: JqdRoutines.pm,v 1.10 2008/12/15 22:06:50 mavrik Exp $
#
######################################################################
#
# Copyright 2007-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Common Job Queue Directory routines
#
######################################################################

package WebJob::JqdRoutines;

require Exporter;

use 5.008;
use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

use Fcntl qw(:DEFAULT :flock);
use FileHandle;
use File::Path;
use WebJob::FdaRoutines;
use WebJob::KvpRoutines;
use WebJob::ValidationRoutines;

@EXPORT = qw(JqdCheckQueueTree JqdCreateQueue JqdDeleteQueue JqdFreezeQueue JqdGetQueuedJobs JqdGetQueueNid JqdGetQueueName JqdIsQueueFrozen JqdLockFile JqdRebuildGroup JqdResolveQueueList JqdSetupQueueMaps JqdThawQueue JqdUnlockFile);

@EXPORT_OK = ();
@ISA = qw(Exporter);
$VERSION = do { my @r = (q$Revision: 1.10 $ =~ /(\d+)/g); sprintf("%d."."%03d" x $#r, @r); };

######################################################################
#
# JqdCheckQueueTree
#
######################################################################

sub JqdCheckQueueTree
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'Directory',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Make sure that the directories required to support the specified
  # queue exist and are writable.
  #
  ####################################################################

  my ($sStateDirectory);

  if (!-d $$phPArgs{'Directory'} || !-W _)
  {
    $$phPArgs{'Error'} = "Queue directory ($$phPArgs{'Directory'}) does not exist or is not writable.";
    return undef;
  }

  foreach my $sState ("hold", "todo", "sent", "open", "done", "pass", "fail", "foul")
  {
    $sStateDirectory = $$phPArgs{'Directory'} . "/" . $sState;
    if (!-d $sStateDirectory || !-W _)
    {
      $$phPArgs{'Error'} = "Queue state directory ($sStateDirectory) does not exist or is not writable.";
      return undef;
    }
  }

  1;
}


######################################################################
#
# JqdCreateQueue
#
######################################################################

sub JqdCreateQueue
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
#     'BeQuiet', # Optional
      'JobQueueDirectory',
      'Queue',
      'QueueGroup',
      'QueueOwner',
      'QueuePermissions',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Define server-side directory structure.
  #
  ####################################################################

  my $sQueueDirectory = $$phPArgs{'JobQueueDirectory'} . "/" . $$phPArgs{'Queue'};
  my $sUid = FdaOwnerToUid({ 'Owner' => $$phPArgs{'QueueOwner'} });
  my $sGid = FdaGroupToGid({ 'Group' => $$phPArgs{'QueueGroup'} });
  my $sMode = FdaPermissionsToMode({ 'Permissions' => $$phPArgs{'QueuePermissions'}, 'Umask' => umask() });

  my $paaPdmugs = # Prefix, Directory, Mode, UID, GID
  [
    [
      $$phPArgs{'JobQueueDirectory'},
      $$phPArgs{'Queue'},
      $sMode,
      $sUid,
      $sGid,
    ],
    [
      $sQueueDirectory,
      "hold",
      $sMode,
      $sUid,
      $sGid,
    ],
    [
      $sQueueDirectory,
      "todo",
      $sMode,
      $sUid,
      $sGid,
    ],
    [
      $sQueueDirectory,
      "sent",
      $sMode,
      $sUid,
      $sGid,
    ],
    [
      $sQueueDirectory,
      "open",
      $sMode,
      $sUid,
      $sGid,
    ],
    [
      $sQueueDirectory,
      "done",
      $sMode,
      $sUid,
      $sGid,
    ],
    [
      $sQueueDirectory,
      "pass",
      $sMode,
      $sUid,
      $sGid,
    ],
    [
      $sQueueDirectory,
      "fail",
      $sMode,
      $sUid,
      $sGid,
    ],
    [
      $sQueueDirectory,
      "foul",
      $sMode,
      $sUid,
      $sGid,
    ],
  ];

  my $paaPfmugs = # Prefix, File, Mode, UID, GID
  [
    [
      $sQueueDirectory,
      "change.lock",
      ($sMode & 07666), # Remove all execute bits.
      $sUid,
      $sGid,
    ],
  ];


  ####################################################################
  #
  # Create server-side directory structure.
  #
  ####################################################################

  foreach my $paPdmug (@{$paaPdmugs})
  {
    my %hPdmugArgs =
    (
      'Prefix'    => $$paPdmug[0],
      'Directory' => $$paPdmug[1],
      'Mode'      => $$paPdmug[2],
      'Uid'       => $$paPdmug[3],
      'Gid'       => $$paPdmug[4],
      'BeQuiet'   => $$phPArgs{'BeQuiet'} || 0,
    );
    if (!defined(FdaCreateDirectory(\%hPdmugArgs)))
    {
      $$phPArgs{'Error'} = $hPdmugArgs{'Error'};
      return undef;
    }
  }

  ####################################################################
  #
  # Create server-side files.
  #
  ####################################################################

  foreach my $paPfmug (@{$paaPfmugs})
  {
    my %hPfmugArgs =
    (
      'Prefix'    => $$paPfmug[0],
      'File'      => $$paPfmug[1],
      'Mode'      => $$paPfmug[2],
      'Uid'       => $$paPfmug[3],
      'Gid'       => $$paPfmug[4],
      'BeQuiet'   => $$phPArgs{'BeQuiet'} || 0,
    );
    if (!defined(FdaCreateFile(\%hPfmugArgs)))
    {
      $$phPArgs{'Error'} = $hPfmugArgs{'Error'};
      return undef;
    }
  }

  1;
}


######################################################################
#
# JqdDeleteQueue
#
######################################################################

sub JqdDeleteQueue
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
#     'BeQuiet', # Optional
      'JobQueueDirectory',
      'Queue',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Delete server-side directory structure.
  #
  ####################################################################

  my $sQueueDirectory = $$phPArgs{'JobQueueDirectory'} . "/" . $$phPArgs{'Queue'};

#FIXME Consider replacing with an equivalent FDA routine, and remove the 'use File::Path;' above.
  rmtree($sQueueDirectory, ($$phPArgs{'BeQuiet'}) ? 0 : 1, 0);

  1;
}


######################################################################
#
# JqdFreezeQueue
#
######################################################################

sub JqdFreezeQueue
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'Directory',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Lock the queue. This is an exclusive lock.
  #
  ####################################################################

  my (%hQueueLockArgs);

  %hQueueLockArgs =
  (
    'LockFile'  => $$phPArgs{'Directory'} . "/" . "change.lock",
  );

  if (!JqdLockFile(\%hQueueLockArgs))
  {
    $$phPArgs{'Error'} = $hQueueLockArgs{'Error'};
    return undef;
  }

  ####################################################################
  #
  # Freeze the queue.
  #
  ####################################################################

  my ($sFreezeFile);

  $sFreezeFile = $$phPArgs{'Directory'} . "/" . "freeze.lock";
  if (!-f $sFreezeFile)
  {
    if (!open(JQDFH, "> $sFreezeFile"))
    {
      $$phPArgs{'Error'} = "File ($sFreezeFile) could not be created/opened ($!).";
      JqdUnlockFile(\%hQueueLockArgs);
      return undef;
    }
    close(JQDFH);
  }

  ####################################################################
  #
  # Unlock the queue.
  #
  ####################################################################

  JqdUnlockFile(\%hQueueLockArgs);

  1;
}


######################################################################
#
# JqdGetQueuedJobs
#
######################################################################

sub JqdGetQueuedJobs
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'Directory',
      'JobCount',
      'JobFiles',       # Array reference
      'JobState',
      'JobType',
#     'ReturnJobFiles', # Optional
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Determine the type of jobs being requested.
  #
  ####################################################################

  my ($sJobRegex);

  if ($$phPArgs{'JobType'} eq "serial")
  {
    $sJobRegex = qq(s\\d{3,5}_(?:\\d{2}_)?\\d{10}_(?:\\d{6}_)?[\\dA-Fa-f]{5,8});
  }
  elsif ($$phPArgs{'JobType'} eq "parallel")
  {
    $sJobRegex = qq(p\\d{3,5}_(?:\\d{2}_)?\\d{10}_(?:\\d{6}_)?[\\dA-Fa-f]{5,8});
  }
  else
  {
    $sJobRegex = qq([ps]\\d{3,5}_(?:\\d{2}_)?\\d{10}_(?:\\d{6}_)?[\\dA-Fa-f]{5,8});
  }

  ####################################################################
  #
  # Open the state directory and create a list of job files.
  #
  ####################################################################

  my (@aJobFiles, $sStateDirectory);

  $sStateDirectory = $$phPArgs{'Directory'} . "/" . $$phPArgs{'JobState'};

  if (!opendir(DIR, $sStateDirectory))
  {
    $$phPArgs{'Error'} = "Directory ($sStateDirectory) could not be opened ($!).";
    return undef;
  }
  @aJobFiles = map("$sStateDirectory/$_", sort(grep(/^$sJobRegex$/, readdir(DIR))));
  closedir(DIR);

  ####################################################################
  #
  # Conditionally fill the output array, but insert no more than the
  # requested number of job files. The reason this is conditional is
  # that some callers may only want to get a job count, so returning
  # anthing more would be wasted effort.
  #
  ####################################################################

  my ($sFinalJobCount, $sReturnJobFiles);

  $sFinalJobCount = ($$phPArgs{'JobCount'} == 0 || $$phPArgs{'JobCount'} >= scalar(@aJobFiles)) ? scalar(@aJobFiles) : $$phPArgs{'JobCount'};

  $sReturnJobFiles = $$phPArgs{'ReturnJobFiles'} || 0;

  if ($sReturnJobFiles)
  {
    @{$$phPArgs{'JobFiles'}} = splice(@aJobFiles, 0, $sFinalJobCount);
  }

  return $sFinalJobCount;
}


######################################################################
#
# JqdGetQueueName
#
######################################################################

sub JqdGetQueueName
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'ClientId',
#FIXME This is a sub hash, but we really want/need to verify the required keys in that hash.
      'CommonRegexes',
      'JobQueueDirectory',
      'JqdQueueNid',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # If the queue NID is zero, return the client's own queue name.
  #
  ####################################################################

  if ($$phPArgs{'JqdQueueNid'} =~ /^0{3,5}$/)
  {
    return $$phPArgs{'ClientId'};
  }

  ####################################################################
  #
  # Resolve the queue NID and return the associated queue name.
  #
  ####################################################################

  my (%hForwardQueueMap, %hQueueMapArgs, %hReverseQueueMap);

  %hQueueMapArgs =
  (
    'ForwardQueueMap'   => \%hForwardQueueMap,
    'JobQueueDirectory' => $$phPArgs{'JobQueueDirectory'},
    'KeyRegex'          => "(?:" . $$phPArgs{'CommonRegexes'}{'ClientId'} . "|" . $$phPArgs{'CommonRegexes'}{'ClientSuppliedFilename'} . ")",
    'ReverseQueueMap'   => \%hReverseQueueMap,
    'SharedLock'        => 1,
    'ValueRegex'        => "\\d+",
  );
  if (!JqdSetupQueueMaps(\%hQueueMapArgs))
  {
    $$phPArgs{'Error'} = $hQueueMapArgs{'Error'};
    return undef;
  }

  return $hReverseQueueMap{$$phPArgs{'JqdQueueNid'}};
}


######################################################################
#
# JqdGetQueueNid
#
######################################################################

sub JqdGetQueueNid
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'ClientId',
#FIXME This is a sub hash, but we really want/need to verify the required keys in that hash.
      'CommonRegexes',
      'JobQueueDirectory',
      'JqdQueueName',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # If the queue name is the client ID, return the zero queue NID.
  #
  ####################################################################

  if ($$phPArgs{'JqdQueueName'} eq $$phPArgs{'ClientId'})
  {
    return 0;
  }

  ####################################################################
  #
  # Lookup the queue name and return the associated queue NID.
  #
  ####################################################################

  my (%hForwardQueueMap, %hQueueMapArgs, %hReverseQueueMap);

  %hQueueMapArgs =
  (
    'ForwardQueueMap'   => \%hForwardQueueMap,
    'JobQueueDirectory' => $$phPArgs{'JobQueueDirectory'},
    'KeyRegex'          => "(?:" . $$phPArgs{'CommonRegexes'}{'ClientId'} . "|" . $$phPArgs{'CommonRegexes'}{'ClientSuppliedFilename'} . ")",
    'ReverseQueueMap'   => \%hReverseQueueMap,
    'SharedLock'        => 1,
    'ValueRegex'        => "\\d+",
  );
  if (!JqdSetupQueueMaps(\%hQueueMapArgs))
  {
    $$phPArgs{'Error'} = $hQueueMapArgs{'Error'};
    return undef;
  }

  return $hForwardQueueMap{$$phPArgs{'JqdQueueName'}};
}


######################################################################
#
# JqdIsQueueFrozen
#
######################################################################

sub JqdIsQueueFrozen
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'Directory',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Check to see if the queue is frozen. If a given queue is frozen,
  # jobs should not be allowed to transition from the 'todo' to the
  # 'sent' queue states. Currently, other state transitions are not
  # affected by a freeze.
  #
  ####################################################################

  my ($sFreezeFile);

  $sFreezeFile = $$phPArgs{'Directory'} . "/" . "freeze.lock";

  return (-f $sFreezeFile) ? 1 : 0;
}


######################################################################
#
# JqdLockFile
#
######################################################################

sub JqdLockFile
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'LockFile',
#     'LockFlags', # Optional
#     'LockMode',  # Optional
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Assign default values where necessary.
  #
  ####################################################################

  my ($sLockFile, $sLockFlags, $sLockMode);

  $sLockFile = $$phPArgs{'LockFile'};

  $sLockMode = ($$phPArgs{'LockMode'}) ? $$phPArgs{'LockMode'} : "+<";

  $sLockFlags = ($$phPArgs{'LockFlags'}) ? $$phPArgs{'LockFlags'} : LOCK_EX;

  ####################################################################
  #
  # Lock the specified file.
  #
  ####################################################################

  my ($oFileHandle, $sModeFile);

  $oFileHandle = new FileHandle;
  if (!defined($oFileHandle))
  {
    $$phPArgs{'Error'} = "Unable to obtain a file handle.";
    return undef;
  }

  if (!$oFileHandle->open("$sLockMode $sLockFile"))
  {
    $$phPArgs{'Error'} = "File ($sLockFile) could not be opened ($!).";
    return undef;
  }

  if (!flock($oFileHandle, $sLockFlags))
  {
    $$phPArgs{'Error'} = "File ($sLockFile) could not be locked ($!).";
    return undef;
  }

  $$phPArgs{'LockHandle'} = $oFileHandle;

  1;
}


######################################################################
#
# JqdRebuildGroup
#
######################################################################

sub JqdRebuildGroup
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'Group',
      'GroupMap',
#     'Members',       # Optional
#     'RemoveMembers', # Optional
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Pull unique members from the group (if it exists).
  #
  ####################################################################

  my (%hMergeGroup);

  if (exists($$phPArgs{'GroupMap'}{$$phPArgs{'Group'}}))
  {
    foreach my $sMember (split(/,/, $$phPArgs{'GroupMap'}{$$phPArgs{'Group'}}))
    {
      if (!defined($sMember) || length($sMember) < 1 || ($sMember =~ /^%(.+)$/ && !exists($$phPArgs{'GroupMap'}{$1})))
      {
        next; # Ignore undefined members or nonexistent groups.
      }
      $hMergeGroup{$sMember}++;
    }
  }

  ####################################################################
  #
  # Pull unique members from the members list (if it is defined).
  #
  ####################################################################

  my ($sRemoveMembers);

  $sRemoveMembers = $$phPArgs{'RemoveMembers'} || 0;

  if (defined($$phPArgs{'Members'}))
  {
    foreach my $sMember (split(/,/, $$phPArgs{'Members'}))
    {
      if (!defined($sMember) || length($sMember) < 1 || ($sMember =~ /^%(.+)$/ && !exists($$phPArgs{'GroupMap'}{$1})))
      {
        next; # Ignore undefined members or nonexistent groups.
      }
      if ($sRemoveMembers)
      {
        delete($hMergeGroup{$sMember});
      }
      else
      {
        $hMergeGroup{$sMember}++;
      }
    }
  }

  ####################################################################
  #
  # Return a lexically sorted, unique, and delimited list.
  #
  ####################################################################

  return join(",", sort(keys(%hMergeGroup)));
}


######################################################################
#
# JqdResolveQueueList
#
######################################################################

sub JqdResolveQueueList
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'ForwardMap',     # Hash reference (input)
      'GroupMap',       # Hash reference (input)
#     'NoRecursion',    # Optinal (input)
      'QueueList',      # Comma delimited string (input)
      'QueueMap',       # Hash reference (output)
      'RecursionStack', # Array reference (internal use)
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Split the queue list up into items, look for group tokens, and
  # resolve them -- use recursion if necessary and enabled.
  #
  ####################################################################

  %{$$phPArgs{'QueueMap'}} = ();

  if ($$phPArgs{'QueueList'} eq "")
  {
    return 1; # Nothing to resolve.
  }
  if ($$phPArgs{'QueueList'} =~ /^((?:(?:%{1,2})?[\w+.:-]+)(?:,(?:%{1,2})?[\w+.:-]+)*)$/)
  {
    foreach my $sItem (split(/,/, $1))
    {
      if ($sItem =~ /^%([^%]+)$/) # Expand group tokens.
      {
        my $sGroup = $1;
        if (exists($$phPArgs{'GroupMap'}{$sGroup}) && defined($$phPArgs{'GroupMap'}{$sGroup}))
        {
          foreach my $sSpentGroup (@{$$phPArgs{'RecursionStack'}})
          {
            if ($sGroup eq $sSpentGroup)
            {
              $$phPArgs{'Error'} = "Group has a recursion loop (" . join(",", @{$$phPArgs{'RecursionStack'}}, $sGroup) . ").";
              return undef;
            }
          }
          if ($$phPArgs{'NoRecursion'})
          {
            foreach my $sMember (split(/,/, $$phPArgs{'GroupMap'}{$sGroup}))
            {
              $$phPArgs{'QueueMap'}{$sMember} = 0;
            }
          }
          else
          {
            push(@{$$phPArgs{'RecursionStack'}}, $sGroup);
            my %hResolverArgs =
            (
              'ForwardMap'     => $$phPArgs{'ForwardMap'},
              'GroupMap'       => $$phPArgs{'GroupMap'},
              'QueueList'      => $$phPArgs{'GroupMap'}{$sGroup},
              'QueueMap'       => {},
              'RecursionStack' => [],
            );
            if (!JqdResolveQueueList(\%hResolverArgs))
            {
              $$phPArgs{'Error'} = $hResolverArgs{'Error'};
              return undef;
            }
            pop(@{$$phPArgs{'RecursionStack'}});
            foreach my $sQueue (sort(keys(%{$hResolverArgs{'QueueMap'}})))
            {
              $$phPArgs{'QueueMap'}{$sQueue} = (exists($$phPArgs{'ForwardMap'}{$sQueue})) ? $$phPArgs{'ForwardMap'}{$sQueue} : 0;
            }
          }
        }
        else
        {
          $$phPArgs{'Error'} = "Group ($sGroup) is not defined.";
          return undef;
        }
      }
      elsif ($sItem =~ /^%(%[^%]+)$/) # Do not expand escaped group tokens.
      {
        $$phPArgs{'QueueMap'}{$1} = 0;
      }
      else # Everything else is assumed to be a group member.
      {
        $$phPArgs{'QueueMap'}{$sItem} = (exists($$phPArgs{'ForwardMap'}{$sItem})) ? $$phPArgs{'ForwardMap'}{$sItem} : 0;
      }
    }
  }
  else
  {
    if (length($$phPArgs{'QueueList'}))
    {
      $$phPArgs{'Error'} = "The specified list does not pass muster.";
      return undef;
    }
  }

  1;
}


######################################################################
#
# JqdSetupQueueMaps
#
######################################################################

sub JqdSetupQueueMaps
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'ForwardQueueMap', # Hash reference
      'JobQueueDirectory',
      'KeyRegex',
      'ReverseQueueMap', # Hash reference
#     'SharedLock',      # Optional
      'ValueRegex',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Conditionally lock the index with a shared lock.
  #
  ####################################################################

  my (%hIndexLockArgs);

  %hIndexLockArgs =
  (
     'LockFile'  => $$phPArgs{'JobQueueDirectory'} . "/" . "queue.index.lock",
     'LockFlags' => LOCK_SH,
  );

  if ($$phPArgs{'SharedLock'} && !JqdLockFile(\%hIndexLockArgs))
  {
    $$phPArgs{'Error'} = $hIndexLockArgs{'Error'};
    return undef;
  }

  ####################################################################
  #
  # Initialize forward and reverse queue maps. These maps are hashes
  # that contain queue name to queue NID and vice-versa mappings.
  #
  #####################################################################

  %{$$phPArgs{'ForwardQueueMap'}} = (); # Initialize an empty forward map.
  %{$$phPArgs{'ReverseQueueMap'}} = (); # Initialize an empty reverse map.
  %hLArgs =
  (
    'File'           => $$phPArgs{'JobQueueDirectory'} . "/" . "queue.index",
    'Properties'     => $$phPArgs{'ForwardQueueMap'},
    'Template'       => { $$phPArgs{'KeyRegex'} => $$phPArgs{'ValueRegex'} },
    'VerifyValues'   => 1,
  );
  if (!KvpGetKvps(\%hLArgs))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'};
    JqdUnlockFile(\%hIndexLockArgs) if ($$phPArgs{'SharedLock'});
    return undef;
  }
  %{$$phPArgs{'ReverseQueueMap'}} = reverse(%{$$phPArgs{'ForwardQueueMap'}}); # This assumes a 1:1 mapping.

  ####################################################################
  #
  # Conditionally unlock the index.
  #
  ####################################################################

  JqdUnlockFile(\%hIndexLockArgs) if ($$phPArgs{'SharedLock'});

  1;
}


######################################################################
#
# JqdThawQueue
#
######################################################################

sub JqdThawQueue
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'Directory',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Lock the queue. This is an exclusive lock.
  #
  ####################################################################

  my (%hQueueLockArgs);

  %hQueueLockArgs =
  (
    'LockFile'  => $$phPArgs{'Directory'} . "/" . "change.lock",
  );

  if (!JqdLockFile(\%hQueueLockArgs))
  {
    $$phPArgs{'Error'} = $hQueueLockArgs{'Error'};
    return undef;
  }

  ####################################################################
  #
  # Unfreeze the queue.
  #
  ####################################################################

  my ($sFreezeFile);

  $sFreezeFile = $$phPArgs{'Directory'} . "/" . "freeze.lock";
  if (-f $sFreezeFile)
  {
    if (!unlink($sFreezeFile))
    {
      $$phPArgs{'Error'} = "File ($sFreezeFile) could not be removed ($!).";
      JqdUnlockFile(\%hQueueLockArgs);
      return undef;
    }
  }

  ####################################################################
  #
  # Unlock the queue.
  #
  ####################################################################

  JqdUnlockFile(\%hQueueLockArgs);

  1;
}


######################################################################
#
# JqdUnlockFile
#
######################################################################

sub JqdUnlockFile
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'LockFile',
      'LockHandle',
#     'LockRemove', # Optional
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Unlock, close, and conditionally remove the lock file.
  #
  ####################################################################

  flock($$phPArgs{'LockHandle'}, LOCK_UN);
  undef($$phPArgs{'LockHandle'}); # According to the documentation, this will automatically close the file.
  if ($$phPArgs{'LockRemove'})
  {
    if (!unlink($$phPArgs{'LockFile'}))
    {
      $$phPArgs{'Error'} = "File ($$phPArgs{'LockFile'}) could not be removed ($!).";
      return undef;
    }
  }

  1;
}


1;

__END__

=pod

=head1 NAME

WebJob::JqdRoutines - Common Job Queue Directory routines

=head1 SYNOPSIS

    use Fcntl qw(:DEFAULT :flock);
    use WebJob::JqdRoutines;

=head1 DESCRIPTION

This module is a collection of common routines designed to support
various WebJob server-side utilities. As such, minimal effort was put
into supporting this code for general consumption. In other words, use
at your own risk and don't expect the interface to remain the same or
backwards compatible from release to release. This module does not
provide an OO interface, nor will it do so anytime soon.

=head1 AUTHOR

Klayton Monroe

=head1 LICENSE

All documentation and code are distributed under same terms and
conditions as B<WebJob>.

=cut
