######################################################################
#
# $Id: KvpRoutines.pm,v 1.11 2008/05/19 14:45:43 mavrik Exp $
#
######################################################################
#
# Copyright 2004-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Key/Value Pair routines
#
######################################################################

package WebJob::KvpRoutines;

require Exporter;

use 5.008;
use strict;
use FileHandle;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
use WebJob::ValidationRoutines;

@EXPORT = qw(KvpGetKvps KvpSetKvps);

@EXPORT_OK = ();
@ISA = qw(Exporter);
$VERSION = do { my @r = (q$Revision: 1.11 $ =~ /(\d+)/g); sprintf("%d."."%03d" x $#r, @r); };

######################################################################
#
# KvpGetKvps
#
######################################################################

sub KvpGetKvps
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
      'File',
#     'MatchKeyCase'   # Optional
      'Properties',
#     'RequireAllKeys' # Optional
      'Template',
#     'VerifyValues',  # Optional
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Open the properties file for read access.
  #
  ####################################################################

  my $sFileHandle;

  my $sOpenMode = "<";

  if ($$phPArgs{'File'} eq "-")
  {
    $sFileHandle = \*STDIN;
  }
  else
  {
    if (!open(KVPFH, "$sOpenMode $$phPArgs{'File'}"))
    {
      $$phPArgs{'Error'} = "File ($$phPArgs{'File'}) could not be opened ($!).";
      return undef;
    }
    $sFileHandle = \*KVPFH;
  }

  ####################################################################
  #
  # Read the properties file. Ignore lines that don't adhere to the
  # key/value pair format. Conditionally ignore case when evaluating
  # keys. If the Template sub-hash is empty, then no key/value pairs
  # will be returned. If strict mode is enabled, every value must
  # pass a regular expression check.
  #
  # Notes for regular expression keys: If the actual key and the
  # corresponding template key don't match when compared as lowercase
  # strings, assume that the template key was specified as a regular
  # expression, and use the actual key (as spelled) to populate the
  # properties hash. This is useful for creating compact templates
  # where the keys vary and generally won't be known ahead of time.
  # This feature, however, may not work as expected if the caller has
  # set the RequireAllKeys option because there is no way to predict
  # which or how many keys will be in the config file. Also, the
  # MatchKeyCase option should be set so that the regular expression
  # can handle case matching on its own.
  #
  ####################################################################

  my $sKeyCaseRegex = ($$phPArgs{'MatchKeyCase'}) ? "" : "(?i)";

  my $sVerifyValues = $$phPArgs{'VerifyValues'} || 0;

  for (my $sLineNumber = 1; my $sLine = <$sFileHandle>; $sLineNumber++)
  {
    $sLine =~ s/[\r\n]+$//; # Remove CRs and LFs.
    $sLine =~ s/([^\\])#.*$/$1/; # Remove comments, but only if they aren't escaped.
    $sLine =~ s/[\\](#)/$1/g; # Remove the escapes that were protecting the comments.
    if (my ($sKey, $sValue) = ($sLine =~ /^([^=]+)=(.*)$/))
    {
      $sKey =~ s/^\s+//; # Remove leading whitespace.
      $sKey =~ s/\s+$//; # Remove trailing whitespace.
      $sValue =~ s/^\s+//; # Remove leading whitespace.
      $sValue =~ s/\s+$//; # Remove trailing whitespace.
      if (defined($sKey) && length($sKey))
      {
        foreach my $sTemplateKey (keys(%{$$phPArgs{'Template'}}))
        {
          if ($sKey =~ /^$sKeyCaseRegex$sTemplateKey$/)
          {
            if ($sVerifyValues && $sValue !~ /^$$phPArgs{'Template'}{$sTemplateKey}$/)
            {
              $$phPArgs{'Error'} = "The value for the \"$sKey\" property on line $sLineNumber does not pass muster.";
              close($sFileHandle) unless ($$phPArgs{'File'} eq "-");
              return undef;
            }
            $$phPArgs{'Properties'}{((lc($sKey) eq lc($sTemplateKey)) ? $sTemplateKey : $sKey)} = $sValue;
          }
        }
      }
    }
  }

  ####################################################################
  #
  # Close the properties file.
  #
  ####################################################################

  close($sFileHandle) unless ($$phPArgs{'File'} eq "-");

  ####################################################################
  #
  # Conditionally verify that all keys in the template are present.
  #
  ####################################################################

  my $sRequireAllKeys = $$phPArgs{'RequireAllKeys'} || 0;

  my @aKeys = ();

  if ($sRequireAllKeys)
  {
    foreach my $sTemplateKey (keys(%{$$phPArgs{'Template'}}))
    {
      if (!exists($$phPArgs{'Properties'}{$sTemplateKey}) || !defined($$phPArgs{'Properties'}{$sTemplateKey}))
      {
        push(@aKeys, $sTemplateKey);
      }
    }
    if (scalar(@aKeys) > 0)
    {
      my $sKeys = join(",", @aKeys);
      $$phPArgs{'Error'} = "One or more required properties ($sKeys) are missing or undefined.";
      return undef;
    }
  }

  1;
}


######################################################################
#
# KvpSetKvps
#
######################################################################

sub KvpSetKvps
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
#     'AppendToFile'   # Optional
      'File',
#     'Delimiter'      # Optional
#     'MatchKeyCase'   # Optional
#     'Newline'        # Optional
      'Properties',
#     'Template',      # Optional
#     'VerifyValues',  # Optional
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  my ($sFile, $phProperties, $phTemplate, $phOptions) = @_;

  ####################################################################
  #
  # Open the properties file for write access.
  #
  ####################################################################

  my $sFileHandle;

  my $sOpenMode = ($$phPArgs{'AppendToFile'}) ? ">>" : ">";

  if ($$phPArgs{'File'} eq "-")
  {
    $sFileHandle = \*STDOUT;
  }
  else
  {
    if (!open(KVPFH, "$sOpenMode $$phPArgs{'File'}"))
    {
      $$phPArgs{'Error'} = "File ($$phPArgs{'File'}) could not be opened ($!).";
      return undef;
    }
    $sFileHandle = \*KVPFH;
  }

  ####################################################################
  #
  # Generate the contents of the Properties hash. If the Template
  # hash exists, only output properties that are common to both
  # hashes. If the VerifyValues option is enabled, test each value to
  # ensure that it's valid according to the specified expression.
  #
  ####################################################################

  my $sDelimiter = $$phPArgs{'Delimiter'} || "=";

  my $sKeyCaseRegex = ($$phPArgs{'MatchKeyCase'}) ? "" : "(?i)";

  my $sNewline = $$phPArgs{'Newline'} || "$/";

  my $sVerifyValues = $$phPArgs{'VerifyValues'} || 0;

  foreach my $sKey (sort(keys(%{$$phPArgs{'Properties'}})))
  {
    my $sValue = $$phPArgs{'Properties'}{$sKey};
    if (exists($$phPArgs{'Template'}))
    {
      my $sHaveKvp = 0;
      foreach my $sTemplateKey (keys(%{$$phPArgs{'Template'}}))
      {
        if ($sKey =~ /^$sKeyCaseRegex$sTemplateKey$/)
        {
          if ($sVerifyValues && $sValue !~ /^$$phPArgs{'Template'}{$sTemplateKey}$/)
          {
            $$phPArgs{'Error'} = "The value for the \"$sKey\" property does not pass muster.";
            close($sFileHandle) unless ($$phPArgs{'File'} eq "-");
            return undef;
          }
          $sKey = $sTemplateKey if (lc($sKey) eq lc($sTemplateKey));
          $sHaveKvp = 1;
          last;
        }
      }
      next unless ($sHaveKvp);
    }
    $sValue =~ s/(#)/\\$1/g; # Insert escapes to protect embedded '#'s.
    if (!print($sFileHandle $sKey, $sDelimiter, $sValue, $sNewline))
    {
      $$phPArgs{'Error'} = "Unable to write the key/value pair for the \"$sKey\" property to $$phPArgs{'File'} ($!).";
      close($sFileHandle) unless ($$phPArgs{'File'} eq "-");
      return undef;
    }
  }

  ####################################################################
  #
  # Close the properties file.
  #
  ####################################################################

  close($sFileHandle) unless ($$phPArgs{'File'} eq "-");

  1;
}

1;

__END__

=pod

=head1 NAME

WebJob::KvpRoutines - Key/Value Pair routines

=head1 SYNOPSIS

    use WebJob::KvpRoutines;

=head1 DESCRIPTION

This module is a collection of key/value pair routines designed to
support various WebJob server-side utilities. As such, minimal effort
was put into supporting this code for general consumption. In other
words, use at your own risk and don't expect the interface to remain
the same or backwards compatible from release to release. This module
does not provide an OO interface, nor will it do so anytime soon.

=head1 AUTHOR

Klayton Monroe

=head1 LICENSE

All documentation and code are distributed under same terms and
conditions as B<WebJob>.

=cut
