######################################################################
#
# $Id: LogRoutines.pm,v 1.5 2008/10/24 05:02:46 mavrik Exp $
#
######################################################################
#
# Copyright 2008-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Log routines
#
######################################################################

package WebJob::LogRoutines;

require Exporter;

use 5.008;
use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

use Fcntl qw(:DEFAULT :flock);
use FileHandle;
use WebJob::TimeRoutines;
use WebJob::ValidationRoutines;

@EXPORT = qw(LogNf1vMessage);

@EXPORT_OK = ();
@ISA = qw(Exporter);
$VERSION = do { my @r = (q$Revision: 1.5 $ =~ /(\d+)/g); sprintf("%d."."%03d" x $#r, @r); };

######################################################################
#
# LogNf1vMessage
#
######################################################################

sub LogNf1vMessage
{
  my ($phPArgs) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phPArgs,
    'Keys' =>
    [
#     'LogEpoch',      # Optional
      'LogFields',
      'LogFile',
      'LogValues',
#     'Newline'        # Optional
#     'RevertToStderr, # Optional
#     'UseGmt',        # Optional
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$phPArgs{'Error'} = $hLArgs{'Error'} if (defined($phPArgs));
    return undef;
  }

  ####################################################################
  #
  # Assign default values where necessary.
  #
  ####################################################################

  my ($sLogEpoch, $sNewline, $sRevertToStderr, $sUseGmt);

  if (exists($$phPArgs{'LogEpoch'}) && defined($$phPArgs{'LogEpoch'}) &&  $$phPArgs{'LogEpoch'} =~ /^\d+$/)
  {
    $sLogEpoch = $$phPArgs{'LogEpoch'};
  }
  else
  {
    $sLogEpoch = time();
  }

  $sNewline = ($$phPArgs{'Newline'}) ? $$phPArgs{'Newline'} : "\n";

  $sRevertToStderr = ($$phPArgs{'RevertToStderr'}) ? 1 : 0;

  $sUseGmt = ($$phPArgs{'UseGmt'}) ? 1 : 0;

  ####################################################################
  #
  # Create date/time stamp.
  #
  ####################################################################

  my $sDateTime = SecondsToDateTime($sLogEpoch, $sUseGmt);

  ####################################################################
  #
  # Construct the log message. Since this routine requires N fixed
  # fields and 1 variable field, the fixed fields must be encoded. If
  # this isn't done, a utility expecting to find N fields could have
  # issues parsing a record that actually contains M fields (where M
  # != N). Additionally, all non-printable characters should be
  # encoded as a general safety measure. Note that the variable field
  # is allowed to have spaces for readability. The encoding scheme
  # used below is a form of URL encoding.
  #
  ####################################################################

  my (@aLogValues, $sLogMessage);

  foreach my $sField (@{$$phPArgs{'LogFields'}})
  {
    my $sLogValue = $$phPArgs{'LogValues'}{$sField};
    if ($sField =~ /^Message$/)
    {
      $sLogValue = (!defined($sLogValue) || length($sLogValue) == 0) ? "--" : "-- $sLogValue";
      $sLogValue =~ s/([\x00-\x1f\x7f-\xff%+])/sprintf("%%%02x", unpack('C',$1))/seg;
      push(@aLogValues, $sLogValue);
    }
    else
    {
      $sLogValue = "-" if (!defined($sLogValue) || length($sLogValue) == 0);
      $sLogValue =~ s/([\x00-\x1f\x7f-\xff%+])/sprintf("%%%02x", unpack('C',$1))/seg;
      $sLogValue =~ s/ /+/sg;
      push(@aLogValues, $sLogValue);
    }
  }
  $sLogMessage = join(" ", $sDateTime, @aLogValues);

  ####################################################################
  #
  # Open (in binary mode) and lock (exclusive) the log file.
  #
  ####################################################################

  my $oFileHandle = new FileHandle;
  if (!defined($oFileHandle))
  {
    if ($sRevertToStderr)
    {
      print STDERR $sLogMessage, $sNewline;
    }
    $$phPArgs{'Error'} = "Unable to obtain a file handle.";
    return undef;
  }

  if (!$oFileHandle->open(">> $$phPArgs{'LogFile'}"))
  {
    if ($sRevertToStderr)
    {
      print STDERR $sLogMessage, $sNewline;
    }
    $$phPArgs{'Error'} = "File ($$phPArgs{'LogFile'}) could not be opened ($!).";
    return undef;
  }
  binmode($oFileHandle);

  if (!flock($oFileHandle, LOCK_EX))
  {
    if ($sRevertToStderr)
    {
      print STDERR $sLogMessage, $sNewline;
    }
    close($oFileHandle);
    $$phPArgs{'Error'} = "File ($$phPArgs{'LogFile'}) could not be locked ($!).";
    return undef;
  }

  ####################################################################
  #
  # Deliver log message.
  #
  ####################################################################

  print $oFileHandle $sLogMessage, $sNewline;

  ####################################################################
  #
  # Cleanup and go home.
  #
  ####################################################################

  flock($oFileHandle, LOCK_UN);
  close($oFileHandle);

  1;
}


1;

__END__

=pod

=head1 NAME

WebJob::LogRoutines - Log routines

=head1 SYNOPSIS

    use WebJob::LogRoutines;

=head1 DESCRIPTION

This module is a collection of log pair routines designed to support
various WebJob server-side utilities. As such, minimal effort was put
into supporting this code for general consumption. In other words, use
at your own risk and don't expect the interface to remain the same or
backwards compatible from release to release. This module does not
provide an OO interface, nor will it do so anytime soon.

=head1 AUTHOR

Klayton Monroe

=head1 LICENSE

All documentation and code are distributed under same terms and
conditions as B<WebJob>.

=cut
