######################################################################
#
# $Id: TimeRoutines.pm,v 1.8 2008/08/06 01:28:05 mavrik Exp $
#
######################################################################
#
# Copyright 2007-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Time-related routines
#
######################################################################

package WebJob::TimeRoutines;

require Exporter;

use 5.008;
use strict;
use Time::Local;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

@EXPORT = qw(DateTimeToSeconds SecondsToDateTime SecondsToDhms);
@EXPORT_OK = ();
@ISA = qw(Exporter);
$VERSION = do { my @r = (q$Revision: 1.8 $ =~ /(\d+)/g); sprintf("%d."."%03d" x $#r, @r); };

######################################################################
#
# DateTimeToSeconds
#
######################################################################

sub DateTimeToSeconds
{
  my ($sDateTime, $sUseGmt) = @_;

  ####################################################################
  #
  # Break the date/time value into pieces. Then, convert and return
  # the number of seconds since the UNIX epoch.
  #
  #     Year = $1
  #    Month = $2
  # MonthDay = $3
  #    Hours = $4
  #  Minutes = $5
  #  Seconds = $6
  #
  ####################################################################

  if (defined($sDateTime) && $sDateTime =~ /^(\d{4})-(\d{2})-(\d{2}) (\d{2}):(\d{2}):(\d{2})$/)
  {
    if ($sUseGmt)
    {
      return timegm($6, $5, $4, $3, $2 - 1, $1 - 1900);
    }
    else
    {
      return timelocal($6, $5, $4, $3, $2 - 1, $1 - 1900);
    }
  }

  return undef;
}


######################################################################
#
# SecondsToDateTime
#
######################################################################

sub SecondsToDateTime
{
  my ($sSeconds, $sUseGmt) = @_;

  if (defined($sSeconds) && $sSeconds =~ /^\d+$/)
  {
    my
    (
      $sSecond,
      $sMinute,
      $sHour,
      $sMonthDay,
      $sMonth,
      $sYear,
      $sWeekDay,
      $sYearDay,
      $sDaylightSavings
    ) = ($sUseGmt) ? gmtime($sSeconds) : localtime($sSeconds);

    my $sDateTime = sprintf("%04s-%02s-%02s %02s:%02s:%02s",
      $sYear + 1900,
      $sMonth + 1,
      $sMonthDay,
      $sHour,
      $sMinute,
      $sSecond
      );

    return $sDateTime;
  }

  return undef;
}


######################################################################
#
# SecondsToDhms
#
######################################################################

sub SecondsToDhms
{
  my ($sSeconds) = @_;

  ####################################################################
  #
  # A value of -1 indicates that seconds is undefined.
  #
  ####################################################################

  my $sDhms = "";

  if (!defined($sSeconds) || $sSeconds < 0)
  {
    $sDhms = "?";
  }
  elsif ($sSeconds == 0)
  {
    $sDhms = "0s";
  }
  else
  {
    my $sDays     = $sSeconds / 86400;
       $sSeconds -= int($sDays) * 86400;
    my $sHours    = $sSeconds / 3600;
       $sSeconds -= int($sHours) * 3600;
    my $sMinutes  = $sSeconds / 60;
       $sSeconds -= int($sMinutes) * 60;

    $sDhms  = (int($sDays)) ? sprintf("%d%s", int($sDays), "d") : "";
    $sDhms .= (int($sDays) || int($sHours)) ? sprintf("%d%s", int($sHours), "h") : "";
    $sDhms .= (int($sDays) || int($sHours) || int($sMinutes)) ? sprintf("%d%s", int($sMinutes), "m") : "";
    $sDhms .= (int($sDays) || int($sHours) || int($sMinutes) || int($sSeconds)) ? sprintf("%d%s", int($sSeconds), "s") : "";
  }

  return $sDhms;
}

1;

__END__

=pod

=head1 NAME

WebJob::TimeRoutines - Time-related routines

=head1 SYNOPSIS

    use WebJob::TimeRoutines;

=head1 DESCRIPTION

This module is a collection of time-related routines designed to
support various WebJob server-side utilities. As such, minimal effort
was put into supporting this code for general consumption. In other
words, use at your own risk and don't expect the interface to remain
the same or backwards compatible from release to release. This module
does not provide an OO interface, nor will it do so anytime soon.

=head1 AUTHOR

Klayton Monroe

=head1 LICENSE

All documentation and code are distributed under same terms and
conditions as B<WebJob>.

=cut
