#!/usr/bin/perl -w
######################################################################
#
# $Id: nph-config.cgi,v 1.9 2008/12/15 04:41:21 mavrik Exp $
#
######################################################################
#
# Copyright 2008-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Process config file requests.
#
######################################################################

use strict;
use Fcntl qw(:flock);
use FindBin qw($Bin $RealBin); use lib ("$Bin/../../lib/perl5/site_perl", "$RealBin/../../lib/perl5/site_perl", "/usr/local/webjob/lib/perl5/site_perl");
use WebJob::KvpRoutines;
use WebJob::LogRoutines;
use WebJob::ValidationRoutines;

######################################################################
#
# Main Routine
#
######################################################################

  my (%hProperties, %hReturnCodes, $sLocalError);

  %hReturnCodes =
  (
    '200' => "OK",
    '204' => "No Content",
    '206' => "Partial Content",
    '403' => "Forbidden",
    '404' => "Not Found",
    '405' => "Method Not Allowed",
    '450' => "Invalid Query",
    '452' => "Username Undefined",
    '453' => "Username-ClientId Mismatch",
    '454' => "Content-Length Undefined",
    '455' => "Content-Length Exceeds Limit",
    '456' => "Content-Length Mismatch",
    '457' => "File Not Available",
    '458' => "Invalid Protocol",
    '470' => "CommonName Undefined",
    '471' => "CommonName-ClientId Mismatch",
    '500' => "Internal Server Error",
    '503' => "Service Unavailable",
    '550' => "Internal Server Initialization Error",
  );

  ####################################################################
  #
  # Punch in and go to work.
  #
  ####################################################################

  $hProperties{'StartTime'} = time();
  $hProperties{'Version'} = sprintf("%s %s", __FILE__, ('$Revision: 1.9 $' =~ /^.Revision: ([\d.]+)/));

  ####################################################################
  #
  # Create/Verify run time environment, and process GET/PUT requests.
  #
  ####################################################################

  if (!defined(CreateRunTimeEnvironment(\%hProperties, \$sLocalError)))
  {
    $hProperties{'ReturnStatus'} = 550;
    $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
    $hProperties{'ErrorMessage'} = $sLocalError;
  }
  else
  {
    if (Yes($hProperties{'SslRequireSsl'}) && (!defined($hProperties{'Https'}) || $hProperties{'Https'} !~ /^[Oo][Nn]$/))
    {
      $hProperties{'ReturnStatus'} = 458;
      $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
      $hProperties{'ErrorMessage'} = "HTTPS required, but client is speaking HTTP";
    }
    else
    {
      if ($hProperties{'RequestMethod'} eq "GET")
      {
        $hProperties{'ReturnStatus'} = ProcessGetRequest(\%hProperties, \$sLocalError);
        $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
        $hProperties{'ErrorMessage'} = $sLocalError;
      }
      else
      {
        $hProperties{'ReturnStatus'} = 405;
        $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
        $hProperties{'ErrorMessage'} = "Method ($hProperties{'RequestMethod'}) not allowed";
      }
    }
  }
  $hProperties{'ServerContentLength'} = SendResponse(\%hProperties);

  ####################################################################
  #
  # Log the transaction.
  #
  ####################################################################

  NphLogMessage(\%hProperties);

  ####################################################################
  #
  # Clean up and go home.
  #
  ####################################################################

  1;


######################################################################
#
# CheckHostAccessList
#
######################################################################

sub CheckHostAccessList
{
  my ($sCidrList, $sIp) = @_;

  my $sCidrRegex = qq(((?:\\d{1,3}[.]){3}(?:\\d{1,3}))\/(\\d{1,2}));

  my $sIpRegex   = qq((?:\\d{1,3}[.]){3}(?:\\d{1,3}));

  if (!defined($sCidrList) || !defined($sIp))
  {
    return 0;
  }

  if ($sIp !~ /^$sIpRegex$/)
  {
    return 0;
  }

  $sCidrList =~ s/\s+//g; # Remove whitespace.

  foreach my $sCidr (split(/,/, $sCidrList))
  {
    next if ($sCidr !~ /^$sCidrRegex$/);
    my ($sNetwork, $sMaskBits) = ($1, $2);
    my $sBits2Clear = 32 - $sMaskBits - 1;
    my $sNetMask = 0xffffffff;
    foreach my $sBit (0..$sBits2Clear)
    {
      $sNetMask ^= 1 << $sBit;
    }
    my $sBinNetwork = hex(sprintf("%02x%02x%02x%02x", split(/\./, $sNetwork)));
    my $sBinNetmask = hex(sprintf("%08x", $sNetMask));
    $sBinNetwork &= $sBinNetmask; # Cleanup the network address -- in case the user did not.
    my $sBinIp = hex(sprintf("%02x%02x%02x%02x", split(/\./, $sIp)));
    if (($sBinIp & $sBinNetmask) == $sBinNetwork)
    {
      return 1; # We have a winner!
    }
  }

  return 0;
}


######################################################################
#
# CreateRunTimeEnvironment
#
######################################################################

sub CreateRunTimeEnvironment
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Put input/output streams in binary mode.
  #
  ####################################################################

  foreach my $sHandle (\*STDIN, \*STDOUT, \*STDERR)
  {
    binmode($sHandle);
  }

  ####################################################################
  #
  # Initialize regex variables.
  #
  ####################################################################

  my %hCommonRegexes =
  (
    'AnyValue'               => qq(.*),
    'AnyValueExceptNothing'  => qq(.+),
    'Base64'                 => qq([0-9A-Za-z+\/]+={0,2}),
    'BaseDirectory'          => qq((?:[A-Za-z]:)?/[\\w./-]+),
    'ClientId'               => qq((?:[A-Za-z](?:(?:[0-9A-Za-z]|[_-](?=[^.]))){0,62})(?:[.][A-Za-z](?:(?:[0-9A-Za-z]|[_-](?=[^.]))){0,62}){0,127}),
    'Filename'               => qq([\\w+.:-]{1,1024}),
    'Decimal16Bit'           => qq(\\d{1,5}),  # 65535
    'Decimal32Bit'           => qq(\\d{1,10}), # 4294967295
    'Decimal64Bit'           => qq(\\d{1,20}), # 18446744073709551615
    'FilePath'               => qq((?:[A-Za-z]:)?/[\\w./-]+),
    'Ip'                     => qq((?:\\d{1,3}\\.){3}\\d{1,3}),
    'YesNo'                  => qq([YyNn]),
    'ProcessId'              => qq(\\d{5}),
    'ServerSuppliedPath'     => qq([\\w+./:\\\\-]+),
    'strftime_Y'             => qq(\\d{4}),
    'strftime_m'             => qq(\\d{2}),
    'strftime_d'             => qq(\\d{2}),
    'strftime_H'             => qq(\\d{2}),
    'strftime_M'             => qq(\\d{2}),
    'strftime_S'             => qq(\\d{2}),
    'strftime_s'             => qq(\\d{10}),
    'FileSuffix'             => qq([\\w.-]{1,32}),
    'SystemVersion'          => qq([\\w ()+,./:-]{1,64}),
    'WebJobVersion'          => qq(webjob[\\w ().-]{1,64}),
  );
  $$phProperties{'CommonRegexes'} = { %hCommonRegexes };

  my %hCustomRegexes =
  (
    'ClientId' => qq(ClientId=($$phProperties{'CommonRegexes'}{'ClientId'})),
    'Filename' => qq(Filename=($$phProperties{'CommonRegexes'}{'Filename'})),
  );
  $$phProperties{'CustomRegexes'} = { %hCustomRegexes };

  $$phProperties{'CustomRegexes'}{'GetQuery'} =
    $$phProperties{'CustomRegexes'}{'ClientId'} . "&" .
    $$phProperties{'CustomRegexes'}{'Filename'}
    ;

  ####################################################################
  #
  # Initialize environment-specific variables. Pull in SSL-related
  # variables, but only if HTTPS is defined and on.
  #
  ####################################################################

  $$phProperties{'ContentLength'}  = $ENV{'CONTENT_LENGTH'};
  $$phProperties{'Https'}          = $ENV{'HTTPS'};
  $$phProperties{'QueryString'}    = $ENV{'QUERY_STRING'};
  $$phProperties{'RemoteAddress'}  = $ENV{'REMOTE_ADDR'};
  $$phProperties{'RemoteUser'}     = $ENV{'REMOTE_USER'};
  $$phProperties{'RequestMethod'}  = $ENV{'REQUEST_METHOD'};
  $$phProperties{'ServerSoftware'} = $ENV{'SERVER_SOFTWARE'};
  $$phProperties{'PropertiesFile'} = $ENV{'CONFIG_PROPERTIES_FILE'};

  if (defined($hProperties{'Https'}) && $hProperties{'Https'} =~ /^[Oo][Nn]$/)
  {
    $$phProperties{'SslClientSDnCn'} = $ENV{'SSL_CLIENT_S_DN_CN'};
  }

  ####################################################################
  #
  # Initialize platform-specific variables.
  #
  ####################################################################

  if ($^O =~ /MSWin32/i)
  {
    $$phProperties{'OSClass'} = "WINDOWS";
    $$phProperties{'Newline'} = "\r\n";
    $$phProperties{'NullDevice'} = "nul";
  }
  else
  {
    $$phProperties{'OSClass'} = "UNIX";
    $$phProperties{'Newline'} = "\n";
    $$phProperties{'NullDevice'} = "/dev/null";
    umask(022);
  }

  ####################################################################
  #
  # Initialize site-specific variables. Note that the properties
  # listed in the custom template are a subset of those in the
  # global template. All values in the custom template may may be
  # overridden through the use of client- and/or command-specific
  # config files.
  #
  ####################################################################

  my (%hCustomConfigTemplate, %hGlobalConfigTemplate, $sLocalError);

  %hGlobalConfigTemplate = # This is the set of site-wide properties.
  (
    'BaseDirectory'          => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'CapContentLength'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'ConfigDirectory'        => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'EnableHostAccessList'   => $$phProperties{'CommonRegexes'}{'YesNo'},
    'LogfilesDirectory'      => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'MaxContentLength'       => $$phProperties{'CommonRegexes'}{'Decimal64Bit'},
    'ProfilesDirectory'      => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'RequireMatch'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'RequireUser'            => $$phProperties{'CommonRegexes'}{'YesNo'},
    'ServerId'               => $$phProperties{'CommonRegexes'}{'ClientId'},
    'SpoolDirectory'         => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'SslRequireSsl'          => $$phProperties{'CommonRegexes'}{'YesNo'},
    'SslRequireCn'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'SslRequireMatch'        => $$phProperties{'CommonRegexes'}{'YesNo'},
    'UseGmt'                 => $$phProperties{'CommonRegexes'}{'YesNo'},
  );
  $$phProperties{'GlobalConfigTemplate'} = { %hGlobalConfigTemplate };

  %hCustomConfigTemplate = # This is the subset of site-wide properties that can be overridden.
  (
    'CapContentLength'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'MaxContentLength'       => $$phProperties{'CommonRegexes'}{'Decimal64Bit'},
    'RequireMatch'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'RequireUser'            => $$phProperties{'CommonRegexes'}{'YesNo'},
    'SslRequireCn'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'SslRequireMatch'        => $$phProperties{'CommonRegexes'}{'YesNo'},
  );
  $$phProperties{'CustomConfigTemplate'} = { %hCustomConfigTemplate };

  GetGlobalConfigProperties($phProperties, \%hGlobalConfigTemplate, \$sLocalError);

  ####################################################################
  #
  # Initialize derived variables. If key directories weren't defined
  # in the config file, define them relative to BaseDirectory.
  #
  ####################################################################

  foreach my $sName ("Config", "Logfiles", "Profiles", "Spool")
  {
    my $sKey = $sName . "Directory";
    if (!exists($$phProperties{$sKey}) || !defined($$phProperties{$sKey}) || length($$phProperties{$sKey}) < 1)
    {
      $$phProperties{$sKey} = $$phProperties{'BaseDirectory'} . "/" . lc($sName);
    }
  }

  $$phProperties{'NphConfigDirectory'} = $$phProperties{'ConfigDirectory'} . "/" . "nph-config";
  $$phProperties{'LogFile'} = $$phProperties{'LogfilesDirectory'} . "/nph-config.log";

  ####################################################################
  #
  # Conditionally initialize the host access list.
  #
  ####################################################################

  if (Yes($$phProperties{'EnableHostAccessList'}))
  {
    if (!defined(SetupHostAccessList($phProperties, \$sLocalError)))
    {
      $$psError = $sLocalError;
      return undef;
    }
  }

  ####################################################################
  #
  # Verify run time environment.
  #
  ####################################################################

  if (!defined(VerifyRunTimeEnvironment($phProperties, \%hGlobalConfigTemplate, \$sLocalError)))
  {
    $$psError = $sLocalError;
    return undef;
  }

  1;
}


######################################################################
#
# GetGlobalConfigProperties
#
######################################################################

sub GetGlobalConfigProperties
{
  my ($phProperties, $phTemplate, $psError) = @_;

  ####################################################################
  #
  # BaseDirectory is the epicenter of activity.
  #
  ####################################################################

  $$phProperties{'BaseDirectory'} = "/var/webjob";

  ####################################################################
  #
  # CapContentLength forces the script to abort when ContentLength
  # exceeds MaxContentLength.
  #
  ####################################################################

  $$phProperties{'CapContentLength'} = "N"; # [Y|N]

  ####################################################################
  #
  # When active, EnableHostAccessList causes the script to consult
  # a host access list to determine whether or not the client will
  # be granted access based on its IP address. Note: This mechanism
  # will fail closed if it's enabled and the access list is missing
  # or does not contain the necessary client ID to IP address
  # mappings.
  #
  ####################################################################

  $$phProperties{'EnableHostAccessList'} = "N"; # [Y|N]

  ####################################################################
  #
  # MaxContentLength specifies the largest upload in bytes the script
  # will accept. If CapContentLength is disabled, this control has no
  # effect.
  #
  ####################################################################

  $$phProperties{'MaxContentLength'} = 100000000; # 100 MB

  ####################################################################
  #
  # RequireMatch forces the script to abort unless ClientId matches
  # RemoteUser. When this value is disabled, any authenticated user
  # will be allowed to issue requests for a given client. Disabling
  # RequireUser implicitly disables RequireMatch.
  #
  ####################################################################

  $$phProperties{'RequireMatch'} = "Y"; # [Y|N]

  ####################################################################
  #
  # RequireUser forces the script to abort unless RemoteUser has been
  # set.
  #
  ####################################################################

  $$phProperties{'RequireUser'} = "Y"; # [Y|N]

  ####################################################################
  #
  # ServerId specifies the identity assigned to the WebJob server.
  #
  ####################################################################

  $$phProperties{'ServerId'} = "server_1";

  ####################################################################
  #
  # SslRequireCn forces the script to abort unless SslClientSDnCn has
  # been set.  If SslRequireSsl is disabled, this and all other SSL
  # controls are ignored.
  #
  ####################################################################

  $$phProperties{'SslRequireCn'} = "N"; # [Y|N]

  ####################################################################
  #
  # SslRequireMatch forces the script to abort if ClientId does not
  # match SslClientSDnCn. When this control is disabled, access will
  # be governed by RequireMatch.  Disabling SslRequireCn implicitly
  # disables SslRequireMatch.  Also, if SslRequireSsl is disabled,
  # this and all other SSL controls are ignored.  The SslRequireMatch
  # check is performed prior to (not instead of) the RequireMatch
  # check.
  #
  ####################################################################

  $$phProperties{'SslRequireMatch'} = "N"; # [Y|N]

  ####################################################################
  #
  # SslRequireSsl forces the script to abort unless the client is
  # speaking HTTPS.  Disabling SslRequireSsl implicitly disables
  # all SSL-related controls.
  #
  ####################################################################

  $$phProperties{'SslRequireSsl'} = "Y"; # [Y|N]

  ####################################################################
  #
  # When active, UseGmt forces the script to convert all time values
  # to GMT. Otherwise, time values are converted to local time.
  #
  ####################################################################

  $$phProperties{'UseGmt'} = "N"; # [Y|N]

  ####################################################################
  #
  # Pull in any externally defined properties. These properties trump
  # internally defined properties.
  #
  ####################################################################

  my (%hLArgs);

  if (!exists($$phProperties{'PropertiesFile'}) || !defined($$phProperties{'PropertiesFile'}))
  {
    $$phProperties{'PropertiesFile'} = $$phProperties{'BaseDirectory'} . "/config/nph-config/nph-config.cfg";
  }

  %hLArgs =
  (
    'File'           => $$phProperties{'PropertiesFile'},
    'Properties'     => $phProperties,
    'RequireAllKeys' => 0,
    'Template'       => $phTemplate,
    'VerifyValues'   => 0,
  );
  KvpGetKvps(\%hLArgs);

  1;
}


######################################################################
#
# NphLogMessage
#
######################################################################

sub NphLogMessage
{
  my ($phProperties) = @_;

  ####################################################################
  #
  # Marshall log arguments.
  #
  ####################################################################

  my (%hLArgs);

  $hLArgs{'LogEpoch'} = time();

  $hLArgs{'LogFields'} =
  [
    'RemoteUser',
    'RemoteAddress',
    'RequestMethod',
    'ClientId',
    'Filename',
    'ContentLength',
    'ServerContentLength',
    'Duration',
    'ReturnStatus',
    'Message'
  ];

  $hLArgs{'LogValues'} =
  {
    'RemoteUser'          => $$phProperties{'RemoteUser'},
    'RemoteAddress'       => $$phProperties{'RemoteAddress'},
    'RequestMethod'       => $$phProperties{'RequestMethod'},
    'ClientId'            => $$phProperties{'ClientId'},
    'Filename'            => $$phProperties{'Filename'},
    'ContentLength'       => $$phProperties{'ContentLength'},
    'ServerContentLength' => $$phProperties{'ServerContentLength'},
    'Duration'            => $hLArgs{'LogEpoch'} - $$phProperties{'StartTime'},
    'ReturnStatus'        => $$phProperties{'ReturnStatus'},
    'Message'             => $$phProperties{'ErrorMessage'},
  };

  $hLArgs{'LogFile'} = $$phProperties{'LogFile'};

  $hLArgs{'Newline'} = $$phProperties{'Newline'};

  $hLArgs{'RevertToStderr'} = 1;

  $hLArgs{'UseGmt'} = (Yes($$phProperties{'UseGmt'})) ? 1 : 0;

  ####################################################################
  #
  # Deliver log message.
  #
  ####################################################################

  LogNf1vMessage(\%hLArgs);

  1;
}


######################################################################
#
# ProcessGetRequest
#
######################################################################

sub ProcessGetRequest
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Proceed only if QueryString matches the GetQuery expression.
  #
  ####################################################################

  my $sQueryString = URLDecode($$phProperties{'QueryString'});

  if ($sQueryString =~ /^$$phProperties{'CustomRegexes'}{'GetQuery'}$/)
  {
    $$phProperties{'ClientId'} = $1 || "nobody";
    $$phProperties{'Filename'} = $2;

    ##################################################################
    #
    # Conditionally do CommonName and client ID checks.
    #
    ##################################################################

    if (Yes($$phProperties{'SslRequireSsl'}))
    {
      if (Yes($$phProperties{'SslRequireCn'}) && (!defined($$phProperties{'SslClientSDnCn'}) || !length($$phProperties{'SslClientSDnCn'})))
      {
        $$psError = "CommonName is undefined or null";
        return 470;
      }

      if (Yes($$phProperties{'SslRequireCn'}) && Yes($$phProperties{'SslRequireMatch'}) && $$phProperties{'SslClientSDnCn'} ne $$phProperties{'ClientId'})
      {
        $$psError = "CommonName ($$phProperties{'SslClientSDnCn'}) does not match client ID ($$phProperties{'ClientId'})";
        return 471;
      }
    }

    ##################################################################
    #
    # Do username and client ID checks.
    #
    ##################################################################

    if (Yes($$phProperties{'RequireUser'}) && (!defined($$phProperties{'RemoteUser'}) || !length($$phProperties{'RemoteUser'})))
    {
      $$psError = "Remote user is undefined or null";
      return 452;
    }

    if (Yes($$phProperties{'RequireUser'}) && Yes($$phProperties{'RequireMatch'}) && $$phProperties{'RemoteUser'} ne $$phProperties{'ClientId'})
    {
      $$psError = "Remote user ($$phProperties{'RemoteUser'}) does not match client ID ($$phProperties{'ClientId'})";
      return 453;
    }

    ##################################################################
    #
    # Do host access list checks.
    #
    ##################################################################

    if (Yes($$phProperties{'EnableHostAccessList'}) && !CheckHostAccessList($$phProperties{'AccessList'}{$$phProperties{'ClientId'}}, $$phProperties{'RemoteAddress'}))
    {
      $$psError = "The ClientId/Address key/value pair ($$phProperties{'ClientId'}=$$phProperties{'RemoteAddress'}/32) is not on the host access list";
      return 403;
    }

    ##################################################################
    #
    # Do content length checks.
    #
    ##################################################################

    if (!defined($$phProperties{'ContentLength'}) || !length($$phProperties{'ContentLength'}))
    {
      $$psError = "Content length is undefined or null";
      return 454;
    }

    if (Yes($$phProperties{'CapContentLength'}) && $$phProperties{'ContentLength'} > $$phProperties{'MaxContentLength'})
    {
      $$psError = "Content length ($$phProperties{'ContentLength'}) exceeds maximum allowed length ($$phProperties{'MaxContentLength'})";
      return 455;
    }

    ##################################################################
    #
    # Locate the requested file and serve it up.
    #
    ##################################################################

    my $sEffectiveFolderList = $$phProperties{'ClientId'};

    foreach my $sFolder (split(/:/, $sEffectiveFolderList))
    {
      my $sGetFile = $$phProperties{'ProfilesDirectory'} . "/" . $sFolder . "/" . $$phProperties{'Filename'};
      if (-e $sGetFile)
      {
        if (!open(FH, "< $sGetFile"))
        {
          $$psError = "Requested file ($sGetFile) could not be opened ($!)";
          return 457;
        }
        binmode(FH);
        $$phProperties{'ReturnHandle'} = \*FH;
        $$psError = "Success";
        return 200;
      }
    }
    $$psError = "Requested file ($$phProperties{'Filename'}) was not found in effective folder list ($sEffectiveFolderList)";
    return 404;
  }
  else
  {
    $$psError = "Invalid query string ($$phProperties{'QueryString'})";
    return 450;
  }
}


######################################################################
#
# SendResponse
#
######################################################################

sub SendResponse
{
  my ($phProperties) = @_;

  ####################################################################
  #
  # Send response header.
  #
  ####################################################################

  my ($sHandle, $sHeader, $sLength, $sReason, $sServer, $sStatus);

  $sHandle = $$phProperties{'ReturnHandle'};
  $sStatus = $$phProperties{'ReturnStatus'};
  $sReason = $$phProperties{'ReturnReason'};
  $sServer = $$phProperties{'ServerSoftware'} || "unknown";
  $sLength = (defined($sHandle)) ? -s $sHandle : 0;

  $sHeader  = "HTTP/1.1 $sStatus $sReason\r\n";
  $sHeader .= "Server: $sServer\r\n";
  $sHeader .= "Content-Type: application/octet-stream\r\n";
  $sHeader .= "Content-Length: $sLength\r\n";
  $sHeader .= "\r\n";

  syswrite(STDOUT, $sHeader, length($sHeader));

  ####################################################################
  #
  # Send content if any.
  #
  ####################################################################

  if (defined($sHandle))
  {
    SysReadWrite($sHandle, \*STDOUT, $sLength, undef);
    close($sHandle);
  }

  return $sLength;
}


######################################################################
#
# SetupHostAccessList
#
######################################################################

sub SetupHostAccessList
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Initialize the access list. This list is a hash that contains
  # client ID to IP address mappings. If the client passes all other
  # authentication tests, a lookup will be performed to determine if
  # the client's IP address is authorized. If it is, processing
  # continues as usual. Otherwise, the CGI script will return a 403
  # status code to the client, log an error message, and abort. If the
  # access list is empty or there is an error processing its file, all
  # access will be denied (i.e., this mechanism fails closed).
  #
  # The format of the list is key/value pair as follows:
  #
  #   <client_id> = <cidr>[,<cidr>]
  #
  #####################################################################

  %{$$phProperties{'AccessList'}} = (); # Initialize an empty list.

  my $sFile = $$phProperties{'NphConfigDirectory'} . "/" . "nph-config-hosts.access";

  if (open(AH, "< $sFile"))
  {
    while (my $sLine = <AH>)
    {
      $sLine =~ s/[\r\n]+$//; # Remove CRs and LFs.
      $sLine =~ s/#.*$//; # Remove comments.
      if ($sLine !~ /^\s*$/)
      {
        my ($sKey, $sValue) = ($sLine =~ /^([^=]*)=(.*)$/);
        $sKey =~ s/^\s+//; # Remove leading whitespace.
        $sKey =~ s/\s+$//; # Remove trailing whitespace.
        $sValue =~ s/^\s+//; # Remove leading whitespace.
        $sValue =~ s/\s+$//; # Remove trailing whitespace.
        if ($sKey =~ /^$$phProperties{'CommonRegexes'}{'ClientId'}$/)
        {
          $$phProperties{'AccessList'}{$sKey} = $sValue;
        }
      }
    }
    close(AH);
  }
  else
  {
    $$psError = "File ($sFile) could not be opened ($!)" if (defined($psError));
    return undef;
  }
}


######################################################################
#
# SysReadWrite
#
######################################################################

sub SysReadWrite
{
  my ($sReadHandle, $sWriteHandle, $sLength, $psError) = @_;

  ####################################################################
  #
  # Read/Write data, but discard data if write handle is undefined.
  #
  ####################################################################

  my ($sData, $sEOF, $sNRead, $sNProcessed, $sNWritten);

  for ($sEOF = $sNRead = $sNProcessed = 0; !$sEOF && $sLength > 0; $sLength -= $sNRead)
  {
    $sNRead = sysread($sReadHandle, $sData, ($sLength > 0x4000) ? 0x4000 : $sLength);
    if (!defined($sNRead))
    {
      $$psError = "Error reading from input stream ($!)" if (defined($psError));
      return undef;
    }
    elsif ($sNRead == 0)
    {
      $sEOF = 1;
    }
    else
    {
      if (defined($sWriteHandle))
      {
        $sNWritten = syswrite($sWriteHandle, $sData, $sNRead);
        if (!defined($sNWritten))
        {
          $$psError = "Error writing to output stream ($!)" if (defined($psError));
          return undef;
        }
      }
      else
      {
        $sNWritten = $sNRead;
      }
      $sNProcessed += $sNWritten;
    }
  }

  return $sNProcessed;
}


######################################################################
#
# URLDecode
#
######################################################################

sub URLDecode
{
  my ($sData) = @_;

  $sData =~ s/\+/ /sg;
  $sData =~ s/%([0-9a-fA-F]{2})/pack('C', hex($1))/seg;

  return $sData;
}


######################################################################
#
# VerifyRunTimeEnvironment
#
######################################################################

sub VerifyRunTimeEnvironment
{
  my ($phProperties, $phRequiredProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure all required properties are defined and valid.
  #
  ####################################################################

  foreach my $sProperty (keys(%$phRequiredProperties))
  {
    my $sValue = $$phProperties{$sProperty};
    if (!defined($sValue) || $sValue !~ /^$$phRequiredProperties{$sProperty}$/)
    {
      $$psError = "$sProperty property ($sValue) is undefined or invalid";
      return undef;
    }
  }

  ####################################################################
  #
  # Make sure the config directory is readable.
  #
  ####################################################################

  if (!-d $$phProperties{'NphConfigDirectory'} || !-R _)
  {
    $$psError = "Config directory ($$phProperties{'NphConfigDirectory'}) does not exist or is not readable";
    return undef;
  }

  ####################################################################
  #
  # Make sure the logfiles directory is readable.
  #
  ####################################################################

  if (!-d $$phProperties{'LogfilesDirectory'} || !-R _)
  {
    $$psError = "Logfiles directory ($$phProperties{'LogfilesDirectory'}) does not exist or is not readable";
    return undef;
  }

  ####################################################################
  #
  # Make sure the profiles directory is readable.
  #
  ####################################################################

  if (!-d $$phProperties{'ProfilesDirectory'} || !-R _)
  {
    $$psError = "Profiles directory ($$phProperties{'ProfilesDirectory'}) does not exist or is not readable";
    return undef;
  }

  ####################################################################
  #
  # Make sure the spool directory is readable.
  #
  ####################################################################

  if (!-d $$phProperties{'SpoolDirectory'} || !-R _)
  {
    $$psError = "Spool directory ($$phProperties{'SpoolDirectory'}) does not exist or is not readable";
    return undef;
  }

  1;
}


######################################################################
#
# Yes
#
######################################################################

sub Yes
{
  return (defined($_[0]) && $_[0] =~ /^[Yy]$/) ? 1 : 0;
}

__END__

=pod

=head1 NAME

nph-config.cgi - Process config file requests.

=head1 SYNOPSIS

    --- http.conf ---
    ...
    Include /var/webjob/config/apache/webjob.conf
    ...
    --- http.conf ---

    --- /var/webjob/config/apache/webjob.conf ---
    SetEnv CONFIG_PROPERTIES_FILE /var/webjob/config/nph-config/nph-config.cfg

    ScriptAlias /cgi-client/ "/usr/local/webjob/cgi/cgi-client/"

    <Directory "/usr/local/webjob/cgi/cgi-client/">
      AllowOverride AuthConfig
      Options None
      Order allow,deny
      Allow from all
    </Directory>
    --- /var/webjob/config/apache/webjob.conf ---

=head1 DESCRIPTION

This utility implements a Common Gateway Interface (CGI), and its
purpose is to process requests for various configuration files.

The name of this utility must begin with the string "nph-".  This
naming convention, which stands for Non Parsed Headers, informs the
web server that the utility is responsible for generating the entire
HTTP response.  Non Parsed Headers are used by this utility because
they allow it to take full advantage of HTTP status code extensions.
This simplifies client-side error checking without giving up the
ability to generate application-specific responses.

The nph-config.cfg config file is used to control the CGI's runtime
behavior. To make use of it, set CONFIG_PROPERTIES_FILE as
appropriate. Upon execution, the script looks for and processes the
file specified by this environment variable.

The cgi-client directory is where nph-config.cgi lives by default.
Access to this directory should be strictly controlled, and any
utilities placed in this directory should operate in the same security
context (or realm).  This utility is designed to read/write data
from/to the server's local disk on behalf of potentially hostile
clients.  Therefore, you'll want to maintain a tight rein to avoid any
unwanted results.

=head1 AUTHOR

Klayton Monroe

=head1 SEE ALSO

nph-webjob.cgi(1), webjob(1)

=head1 LICENSE

All documentation and code are distributed under same terms and
conditions as B<WebJob>.

=cut
