/*-
 ***********************************************************************
 *
 * $Id: webjob-dsvtool.c,v 1.3 2008/01/12 17:37:54 mavrik Exp $
 *
 ***********************************************************************
 *
 * Copyright 2006-2008 The WebJob Project, All Rights Reserved.
 *
 ***********************************************************************
 */
#include "all-includes.h"

/*-
 ***********************************************************************
 *
 * Global Variables.
 *
 ***********************************************************************
 */
static DSV_PROPERTIES *gpsProperties;

/*-
 ***********************************************************************
 *
 * DsvMain
 *
 ***********************************************************************
 */
int
main(int iArgumentCount, char *ppcArgumentVector[])
{
  const char          acRoutine[] = "Main()";
  char                acLocalError[MESSAGE_SIZE] = "";
  int                 iError = 0;
  DSV_PROPERTIES     *psProperties = NULL;

  /*-
   *********************************************************************
   *
   * Punch in and go to work.
   *
   *********************************************************************
   */
  iError = DsvBootStrap(acLocalError);
  if (iError != ER_OK)
  {
    fprintf(stderr, "%s: %s: %s\n", PROGRAM_NAME, acRoutine, acLocalError);
    DsvShutdown(XER_BootStrap);
  }
  psProperties = (DSV_PROPERTIES *) DsvGetPropertiesReference();

  /*-
   *********************************************************************
   *
   * Process command line arguments.
   *
   *********************************************************************
   */
  iError = DsvProcessArguments(iArgumentCount, ppcArgumentVector, psProperties, acLocalError);
  if (iError != ER_OK)
  {
    fprintf(stderr, "%s: %s: %s\n", PROGRAM_NAME, acRoutine, acLocalError);
    DsvShutdown(XER_ProcessArguments);
  }

  /*-
   *********************************************************************
   *
   * Answer the mail.
   *
   *********************************************************************
   */
  switch (psProperties->iRunMode)
  {
  case DSV_SIGN:
    iError = DsvSign(psProperties, acLocalError);
    if (iError != ER_OK)
    {
      fprintf(stderr, "%s: %s: %s\n", PROGRAM_NAME, acRoutine, acLocalError);
      DsvShutdown(XER_PayloadSigningError);
    }
    break;
  case DSV_VERIFY:
    iError = DsvVerify(psProperties, acLocalError);
    if (iError != ER_OK)
    {
      if (iError == ER_SignatureVerificationFailed)
      {
        DsvShutdown(XER_SignatureVerificationFailed);
      }
      else
      {
        fprintf(stderr, "%s: %s: %s\n", PROGRAM_NAME, acRoutine, acLocalError);
        DsvShutdown(XER_SignatureVerificationError);
      }
    }
    break;
  default:
    fprintf(stderr, "%s: %s: Invalid run mode (%d). That shouldn't happen!\n", PROGRAM_NAME, acRoutine, psProperties->iRunMode);
    DsvShutdown(XER_RunMode);
    break;
  }

  /*-
   *********************************************************************
   *
   * Shutdown and go home.
   *
   *********************************************************************
   */
  DsvShutdown(XER_OK);

  return XER_OK;
}


/*-
 ***********************************************************************
 *
 * DsvBootStrap
 *
 ***********************************************************************
 */
int
DsvBootStrap(char *pcError)
{
  const char          acRoutine[] = "DsvBootStrap()";
  char                acLocalError[MESSAGE_SIZE] = "";
  DSV_PROPERTIES     *psProperties = NULL;
  unsigned char       aucSeed[DSV_SEED_LENGTH];

  /*-
   *********************************************************************
   *
   * Initialize OpenSSL components.
   *
   * Note: OpenSSL_add_all_algorithms() may be required to prevent the
   *       following error:
   *
   *       error:0906B072:PEM routines:PEM_get_EVP_CIPHER_INFO:unsupported encryption
   *
   *       which occurrs when the program is run in sign mode.
   *
   *********************************************************************
   */
  ERR_load_crypto_strings();
  OpenSSL_add_all_algorithms();
  RAND_seed(DsvGenerateSeed(aucSeed, DSV_SEED_LENGTH), DSV_SEED_LENGTH);

  /*-
   *********************************************************************
   *
   * Allocate and initialize the properties structure.
   *
   *********************************************************************
   */
  psProperties = (DSV_PROPERTIES *) DsvNewProperties(acLocalError);
  if (psProperties == NULL)
  {
    snprintf(pcError, MESSAGE_SIZE, "%s: %s", acRoutine, acLocalError);
    return ER;
  }
  DsvSetPropertiesReference(psProperties);

  return ER_OK;
}


/*-
 ***********************************************************************
 *
 * DsvFreeProperties
 *
 ***********************************************************************
 */
void
DsvFreeProperties(DSV_PROPERTIES *psProperties)
{
  DSV_CERT_NODE      *psTemp = NULL;
  DSV_CERT_NODE      *psFree = NULL;

  if (psProperties != NULL)
  {
    psTemp = psProperties->psCertList;
    while (psTemp != NULL)
    {
      psFree = psTemp;
      psTemp = psTemp->psNext;
      DsvFreeCertNode(psFree);
    }
    free(psProperties);
  }
}


/*-
 ***********************************************************************
 *
 * DsvGetPropertiesReference
 *
 ***********************************************************************
 */
DSV_PROPERTIES *
DsvGetPropertiesReference(void)
{
  return gpsProperties;
}


/*-
 ***********************************************************************
 *
 * DsvNewProperties
 *
 ***********************************************************************
 */
DSV_PROPERTIES *
DsvNewProperties(char *pcError)
{
  const char          acRoutine[] = "DsvNewProperties()";
  DSV_PROPERTIES     *psProperties = NULL;

  psProperties = (DSV_PROPERTIES *) calloc(sizeof(DSV_PROPERTIES), 1);
  if (psProperties == NULL)
  {
    snprintf(pcError, MESSAGE_SIZE, "%s: calloc(): %s", acRoutine, strerror(errno));
    return NULL;
  }
  return psProperties;
}


/*-
 ***********************************************************************
 *
 * DsvProcessArguments
 *
 ***********************************************************************
 */
int
DsvProcessArguments(int iArgumentCount, char *ppcArgumentVector[], DSV_PROPERTIES *psProperties, char *pcError)
{
  int                 iComplete = 0;

  if (iArgumentCount < 2)
  {
    DsvUsage();
  }

  iArgumentCount -= 2;

  if (strcmp(ppcArgumentVector[1], "-V") == 0 || strcmp(ppcArgumentVector[1], "--verify-signature") == 0)
  {
    if (iArgumentCount == 4)
    {
      psProperties->iRunMode = DSV_VERIFY;
      if (strcmp(ppcArgumentVector[2], "-f") == 0 || strcmp(ppcArgumentVector[2], "--cert-file") == 0)
      {
        psProperties->iVerifyMode = DSV_VERIFY_VIA_FILE;
        psProperties->pcCertFile = ppcArgumentVector[3];
        psProperties->pcPayload = ppcArgumentVector[4];
        psProperties->pcSigFile = ppcArgumentVector[5];
        iComplete = 1;
      }
      else if (strcmp(ppcArgumentVector[2], "-t") == 0 || strcmp(ppcArgumentVector[2], "--cert-tree") == 0)
      {
        psProperties->iVerifyMode = DSV_VERIFY_VIA_TREE;
        psProperties->pcCertTree = ppcArgumentVector[3];
        psProperties->pcPayload = ppcArgumentVector[4];
        psProperties->pcSigFile = ppcArgumentVector[5];
        iComplete = 1;
      }
    }
  }
  else if (strcmp(ppcArgumentVector[1], "-s") == 0 || strcmp(ppcArgumentVector[1], "--sign-payload") == 0)
  {
    if (iArgumentCount >= 3)
    {
      psProperties->iRunMode = DSV_SIGN;
      if (strcmp(ppcArgumentVector[2], "-k") == 0 || strcmp(ppcArgumentVector[2], "--key-file") == 0)
      {
        psProperties->pcKeyFile = ppcArgumentVector[3];
        psProperties->ppcPayloadVector = &ppcArgumentVector[4];
        psProperties->iPayloadCount = iArgumentCount - 2;
        iComplete = 1;
      }
    }
  }
  else if (strcmp(ppcArgumentVector[1], "-v") == 0 || strcmp(ppcArgumentVector[1], "--version") == 0)
  {
    DsvVersion();
  }

  if (!iComplete)
  {
    DsvUsage();
  }

  return ER_OK;
}


/*-
 ***********************************************************************
 *
 * DsvSetPropertiesReference
 *
 ***********************************************************************
 */
void
DsvSetPropertiesReference(DSV_PROPERTIES *psProperties)
{
  gpsProperties = psProperties;
}


/*-
 ***********************************************************************
 *
 * DsvShutdown
 *
 ***********************************************************************
 */
void
DsvShutdown(int iError)
{
  DsvFreeProperties(DsvGetPropertiesReference());
  exit(iError);
}


/*-
 ***********************************************************************
 *
 * DsvSign
 *
 ***********************************************************************
 */
int
DsvSign(DSV_PROPERTIES *psProperties, char *pcError)
{
  const char          acRoutine[] = "DsvSign()";
  char                acLocalError[MESSAGE_SIZE] = "";
  int                 iError = 0;
  int                 iErrorCount = 0;
  int                 iIndex = 0;
  DSV_SIGNATURE      *psSignature = NULL;
  EVP_PKEY           *psPrivateKey = NULL;

  /*-
   *********************************************************************
   *
   * Load the private key from the specified file. Then, sign all of
   * the specified payloads. Don't abort on error since there can be
   * more than one payload in any given run.
   *
   *********************************************************************
   */
  psPrivateKey = DsvLoadPrivateKey(psProperties->pcKeyFile, acLocalError);
  if (psPrivateKey == NULL)
  {
    snprintf(pcError, MESSAGE_SIZE, "%s: %s", acRoutine, acLocalError);
    return ER;
  }
  for (iIndex = 0; iIndex < psProperties->iPayloadCount; iIndex++)
  {
    psSignature = DsvNewSignature(acLocalError);
    if (psSignature == NULL)
    {
      fprintf(stderr, "%s: %s: %s\n", PROGRAM_NAME, acRoutine, acLocalError);
      iErrorCount++;
      continue;
    }
    iError = DsvSignPayload(psProperties->ppcPayloadVector[iIndex], psPrivateKey, psSignature, acLocalError);
    if (iError != ER_OK)
    {
      fprintf(stderr, "%s: %s: %s\n", PROGRAM_NAME, acRoutine, acLocalError);
      DsvFreeSignature(psSignature);
      iErrorCount++;
      continue;
    }
    iError = DsvWriteSignature(psProperties->ppcPayloadVector[iIndex], DSV_SIGNATURE_EXTENSION, (char *) psSignature->pucBase64Signature, acLocalError);
    if (iError != ER_OK)
    {
      fprintf(stderr, "%s: %s: %s\n", PROGRAM_NAME, acRoutine, acLocalError);
      DsvFreeSignature(psSignature);
      iErrorCount++;
      continue;
    }
    DsvFreeSignature(psSignature);
  }
  EVP_PKEY_free(psPrivateKey);

  if (iErrorCount)
  {
    snprintf(pcError, MESSAGE_SIZE, "%s: There were %d signing errors.", acRoutine, iErrorCount);
    return ER;
  }

  return ER_OK;
}


/*-
 ***********************************************************************
 *
 * DsvUsage
 *
 ***********************************************************************
 */
void
DsvUsage(void)
{
  fprintf(stderr, "\n");
  fprintf(stderr, "Usage: %s {-s|--sign-payload} {-k|--key-file} key payload [payload ...]\n", PROGRAM_NAME);
  fprintf(stderr, "       %s {-V|--verify-signature} {-f|--cert-file} cert payload signature\n", PROGRAM_NAME);
  fprintf(stderr, "       %s {-V|--verify-signature} {-t|--cert-tree} tree payload signature\n", PROGRAM_NAME);
  fprintf(stderr, "       %s {-v|--version}\n", PROGRAM_NAME);
  fprintf(stderr, "\n");
  exit(XER_Usage);
}


/*-
 ***********************************************************************
 *
 * DsvVerify
 *
 ***********************************************************************
 */
int
DsvVerify(DSV_PROPERTIES *psProperties, char *pcError)
{
  const char          acRoutine[] = "DsvVerify()";
  char                acLocalError[MESSAGE_SIZE] = "";
  int                 iError = 0;
  DSV_SIGNATURE      *psSignature = NULL;

  /*-
   *********************************************************************
   *
   * Load the public keys. If we are in file verify mode, there will
   * only be one key to load. If we are in tree verify mode, attempt
   * to load every file in the specified tree, but do not abort on a
   * load error. In either case, the result should be a cert list.
   *
   *********************************************************************
   */
  switch (psProperties->iVerifyMode)
  {
  case DSV_VERIFY_VIA_FILE:
    psProperties->psCertList = DsvLoadPublicKey(psProperties->pcCertFile, acLocalError);
    if (psProperties->psCertList == NULL)
    {
      snprintf(pcError, MESSAGE_SIZE, "%s: %s", acRoutine, acLocalError);
      return ER;
    }
    break;
  case DSV_VERIFY_VIA_TREE:
    psProperties->psCertList = DsvLoadPublicKeys(psProperties->pcCertTree, acLocalError);
    if (psProperties->psCertList == NULL)
    {
      snprintf(pcError, MESSAGE_SIZE, "%s: %s", acRoutine, acLocalError);
      return ER;
    }
    break;
  default:
    snprintf(pcError, MESSAGE_SIZE, "%s: Invalid verify mode (%d). That shouldn't happen!", acRoutine, psProperties->iVerifyMode);
    return ER;
    break;
  }

  /*-
   *********************************************************************
   *
   * Load the base64-encoded signature from the specified file.
   *
   *********************************************************************
   */
  psSignature = DsvLoadSignature(psProperties->pcSigFile, acLocalError);
  if (psSignature == NULL)
  {
    snprintf(pcError, MESSAGE_SIZE, "%s: %s", acRoutine, acLocalError);
    return ER;
  }

  /*-
   *********************************************************************
   *
   * Begin the verification process by hashing the payload file. This
   * step need only be done once, and if there are multiple keys to
   * check, it saves us from having to rehash the payload each time.
   *
   *********************************************************************
   */
  iError = DsvProcessPayloadFile(psProperties->pcPayload, psSignature, acLocalError);
  if (iError != ER_OK)
  {
    snprintf(pcError, MESSAGE_SIZE, "%s: %s", acRoutine, acLocalError);
    return ER;
  }

  /*-
   *********************************************************************
   *
   * Loop through the keys until a valid signature is found.
   *
   *********************************************************************
   */
  iError = DsvVerifySignature(psProperties->psCertList, psProperties->iVerifyMode, psSignature, acLocalError);
  switch (iError)
  {
  case DSV_SIGNATURE_VERIFICATION_ERROR:
    snprintf(pcError, MESSAGE_SIZE, "%s: %s", acRoutine, acLocalError);
    fprintf(stdout, "Signature verification failed due to errors.\n");
    return ER;
    break;
  case DSV_SIGNATURE_VERIFICATION_PASSED:
    fprintf(stdout, "Signature is valid. CommonName = [%s]\n", psSignature->pcCommonName);
    return ER_OK;
    break;
  case DSV_SIGNATURE_VERIFICATION_FAILED:
    fprintf(stdout, "Signature verification failed.\n");
    return ER_SignatureVerificationFailed;
    break;
  default:
    snprintf(pcError, MESSAGE_SIZE, "%s: Invalid verification status (%d). That shouldn't happen!", acRoutine, iError);
    return ER;
    break;
  }

  return ER_OK;
}


/*-
 ***********************************************************************
 *
 * DsvVersion
 *
 ***********************************************************************
 */
void
DsvVersion(void)
{
  fprintf(stdout, "%s %s %d-bit, %s\n", PROGRAM_NAME, VERSION, (int) (sizeof(&DsvVersion) * 8), SSLeay_version(SSLEAY_VERSION));
  exit(XER_OK);
}
